/* Model for simulation of falling sand.
   Copyright (C) 2006 Carl Fredrik Hammar (hammy.lite@gmail.com)

   This file is part of GP2X Sand.

   GP2X Sand is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   GP2X Sand is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with GP2X Sand; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */

#ifndef SANDBOX_H
#define SANDBOX_H

#include <stdbool.h>
#include <stdint.h>
#include "SDL.h"

#include "globals.h"
#include "action.h"
#include "probability.h"
#include "vmath.h"

/* The sandbox is where all the particles are stored and interact with
   each-other and ti thus the base for this application.  Though the
   analogy to a sandbox doesn't really hold since it isn't palaced
   horizontally, it's such a nice and concise analogy that I decided to
   keep it.

   FIXME: STOP MAKING EXCUSES AND COMMENT ;) */

enum element_enum
{
  NOTHING,
  WALL,
  SAND,
  WATER,
  OIL,
  SALT,
  SALTWATER,
  STEAM,
  PLANT,
  FIRE,
  ELEMENTS_MAX
};

typedef uint32_t particle_t;

typedef struct element_struct element_t;

struct transformation_struct
{
  prob_t probability;
  element_t *result_primary; /* Result to be placed in initiator. */
  element_t *result_secondary; /* Result to be placed in target. */
  bool shuffle_result_flag; /* Randomly switch the results if true. */
};

typedef struct transformation_struct transformation_t;

struct element_struct
{
  uint16_t color;
  uint32_t weight;
  prob_t resistance; /* Probability of resisting displacement. */
  bool active_flag; /* Whether it reacts with other elements. */
  bool mobile_flag; /* Wether it can be moved. */
  bool fall_flag; /* Whether it falls or rises. */
  bool float_flag; /* Whether it neither falls nor rises. */
  transformation_t transformations[ELEMENTS_MAX];
};

/* The sandbox is divided into sectors enabling some code that examines
   or updates the content of the sandbox to skip sectors that are
   homogenous or stable. */
struct sandbox_sector_str
{
  vect_t position;
  vect_t dimensions;

  /* The most common element in this sector. */
  element_t *dominant_element_ptr;

  /* True if all particles are of the same element. */
  bool homogenous_flag;

  /* True if contained particles won't change. */
  bool stable_flag;

  /* True if sector needs to be updated. */
  bool dirty_flag;
};

typedef struct sandbox_sector_str sandbox_sector_t;

extern element_t *const element_nothing_ptr;
extern element_t *const element_wall_ptr;
extern element_t *const element_sand_ptr;
extern element_t *const element_water_ptr;
extern element_t *const element_oil_ptr;
extern element_t *const element_salt_ptr;

void
sandbox_init ();

void
sandbox_sectors_get (sandbox_sector_t **sectors_ptr_dst, int *length_dst);

void
sandbox_emit (int x, int width, particle_t type, float p);

void
sandbox_update ();

void
sandbox_clear ();

void
sandbox_particle_init (vect_t position, element_t *element_ptr);

particle_t
sandbox_particle_get (vect_t position);

void
sandbox_particle_set (vect_t position, particle_t particle);

element_t*
sandbox_element_first ();

element_t*
sandbox_element_last ();

element_t*
sandbox_element_next (element_t *element_ptr);

element_t*
sandbox_element_prev (element_t *element_ptr);

uint16_t
particle_color_get (particle_t particle);

element_t*
particle_element_get (particle_t particle);

extern action_t *const sandbox_action_clear_ptr;

#endif /* not SANDBOX_H */
